#ifndef lint
static const char yysccsid[] = "@(#)yaccpar	1.9 (Berkeley) 02/21/93";
#endif

#include <stdlib.h>

#define YYBYACC 1
#define YYMAJOR 1
#define YYMINOR 9
#define YYPATCH 20050813

#define YYEMPTY (-1)
#define yyclearin    (yychar = YYEMPTY)
#define yyerrok      (yyerrflag = 0)
#define YYRECOVERING (yyerrflag != 0)

extern int yyparse(void);

static int yygrowstack(void);
#define YYPREFIX "yy"
#line 2 "./config.y"
/*
 * BRLTTY - A background process providing access to the console screen (when in
 *          text mode) for a blind person using a refreshable braille display.
 *
 * Copyright (C) 1995-2007 by The BRLTTY Developers.
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation.  Please see the file COPYING for details.
 *
 * Web Page: http://mielke.cc/brltty/
 *
 * This software is maintained by Dave Mielke <dave@mielke.cc>.
 */

/*
 * Support for all Papenmeier Terminal + config file
 *   Heimo.Schn <heimo.schoen@gmx.at>
 *   August Hrandl <august.hoerandl@gmx.at>
 *
 * Papenmeier/read_config.y
 *  read (scan + interpret) the configuration file - included by braille.c
 *  this file can be used as a standalone test programm - see
 *   Makefile for details
 */

#include "prologue.h"

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>

#include "brl-cfg.h"

#define YYERROR_VERBOSE
static int yylex (void);
static int yyerror (char*);    
extern int yyparse (void);

/* to be called: */
static int parse (void);

static char *nameval = NULL;
static int numval, keyindex, cmdval; 

static int numkeys = 0;
static int16_t keys[KEYMAX];

static FILE *configurationFile = NULL;
static int lineNumber = 1;

static int terminalsSize;
static int commandsSize;

static int
reallocateTable (void *reference, int size, int count) {
  void **address = reference;
  void *newAddress = realloc(*address, (count * size));
  if (!count || (newAddress != NULL)) {
    *address = newAddress;
    return 1;
  } else {
    yyerror("insufficient memory");
  }
  return 0;
}

static int
allocateTable (void *reference, int size, int count) {
  void **address = reference;
  *address = NULL;
  return reallocateTable(address, size, count);
}

static int
duplicateString (char **address, const char *string) {
  int count = strlen(string) + 1;
  if (!allocateTable(address, 1, count)) return 0;
  memcpy(*address, string, count);
  return 1;
}

static int
ensureTableSize (void *reference, int size, int count, int *limit, int increment) {
  if (count == *limit) {
    int newLimit = *limit + increment;
    if (!reallocateTable(reference, size, newLimit)) return 0;
    *limit = newLimit;
  }
  return 1;
}

static TerminalDefinition *
getCurrentTerminal (void) {
  if (pmTerminalCount > 0) return &pmTerminals[pmTerminalCount - 1];
  yyerror("terminal not being defined");
  return NULL;
}

static int
finishCurrentTerminal (void) {
  if (pmTerminalCount) {
    TerminalDefinition *terminal = &pmTerminals[pmTerminalCount - 1];
    if (!reallocateTable(&terminal->statusCells, sizeof(*terminal->statusCells), terminal->statusCount)) return 0;
    if (!reallocateTable(&terminal->modifiers, sizeof(*terminal->modifiers), terminal->modifierCount)) return 0;
    if (!reallocateTable(&terminal->commands, sizeof(*terminal->commands), terminal->commandCount)) return 0;
  }
  return 1;
}

static TerminalDefinition *
addTerminal (int identifier) {
  if (finishCurrentTerminal()) {
    if (ensureTableSize(&pmTerminals, sizeof(*pmTerminals),
                        pmTerminalCount, &terminalsSize, 4)) {
      TerminalDefinition *terminal = &pmTerminals[pmTerminalCount++];
      terminal->identifier = identifier;
      terminal->name = NULL;
      terminal->helpFile = NULL;

      terminal->columns = 0;
      terminal->rows = 1;

      terminal->frontKeys = 0;
      terminal->hasBar = 0;
      terminal->leftSwitches = 0;
      terminal->rightSwitches = 0;
      terminal->leftKeys = 0;
      terminal->rightKeys = 0;

      terminal->statusCount = 0;
      terminal->modifierCount = 0;
      terminal->commandCount = 0;

      terminal->statusCells = NULL;
      terminal->modifiers = NULL;
      terminal->commands = NULL;

      commandsSize = 0;
      return terminal;
    }
  }
  return NULL;
}

static int
setName (char *name) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    if (!terminal->name) {
      if (duplicateString(&terminal->name, name)) {
        return 1;
      }
    } else {
      yyerror("duplicate name");
    }
  }
  return 0;
}

static int
setHelp (char *file) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    if (!terminal->helpFile) {
      if (duplicateString(&terminal->helpFile, file)) {
        return 1;
      }
    } else {
      yyerror("duplicate help file");
    }
  }
  return 0;
}

static int
setColumns (int columns) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    if ((0 < columns) && (columns <= 0XFF)) {
      if (!terminal->columns) {
        terminal->columns = columns;
        return 1;
      } else {
        yyerror("duplicate column count");
      }
    } else {
      yyerror("invalid column count");
    }
  }
  return 0;
}

static int
setStatusCells (int count) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    if ((0 < count) && (count <= 0XFF)) {
      if (!terminal->statusCount) {
        if (allocateTable(&terminal->statusCells, sizeof(*terminal->statusCells), count)) {
          {
            int s;
            for (s=0; s<count; s++) terminal->statusCells[s] = OFFS_EMPTY;
          }

          terminal->statusCount = count;
          return 1;
        }
      } else {
        yyerror("duplicate status cell count");
      }
    } else {
      yyerror("invalid status cell count");
    }
  }
  return 0;
}

static int
setFrontKeys (int count) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    if ((0 < count) && (count <= 0XFF)) {
      if (!terminal->frontKeys) {
        terminal->frontKeys = count;
        return 1;
      } else {
        yyerror("duplicate front key count");
      }
    } else {
      yyerror("invalid front key count");
    }
  }
  return 0;
}

static int
setHasBar (int ls, int rs, int lk, int rk) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    if (terminal->hasBar) {
      yyerror("duplicate bar specification");
    } else if ((ls < 0) || (ls > 1)) {
      yyerror("invalid left switches count");
    } else if ((rs < 0) || (rs > 1)) {
      yyerror("invalid right switches count");
    } else if ((lk < 0) || (lk > 1)) {
      yyerror("invalid left keys count");
    } else if ((rk < 0) || (rk > 1)) {
      yyerror("invalid right keys count");
    } else {
      terminal->hasBar = 1;
      terminal->leftSwitches = ls;
      terminal->rightSwitches = rs;
      terminal->leftKeys = lk;
      terminal->rightKeys = rk;
      return 1;
    }
  }
  return 0;
}

static int
setStatusCell (int number, int code) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    if ((0 < number) && (number <= terminal->statusCount)) {
      uint16_t *cell = &terminal->statusCells[number-1];
      if (*cell == OFFS_EMPTY) {
        *cell = code;
        return 1;
      } else {
        yyerror("duplicate status cell");
      }
    } else {
      yyerror("invalid status cell number");
    }
  }
  return 0;
}

static int
addModifier (int key) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    {
      int m;
      for (m=0; m<terminal->modifierCount; m++) {
        if (terminal->modifiers[m] == key) {
          yyerror("duplicate modifier");
          return 0;
        }
      }
    }

    if (terminal->modifierCount < MODMAX) {
      if (terminal->modifiers ||
          allocateTable(&terminal->modifiers, sizeof(*terminal->modifiers), MODMAX)) {
        terminal->modifiers[terminal->modifierCount++] = key;
        return 1;
      }
    } else {
      yyerror("too many modifiers");
    }
  }
  return 0;
}

static CommandDefinition *
addCommand (int code) {
  TerminalDefinition *terminal = getCurrentTerminal();
  if (terminal) {
    if (ensureTableSize(&terminal->commands, sizeof(*terminal->commands),
                        terminal->commandCount, &commandsSize, 0X20)) {
      CommandDefinition *cmd = &terminal->commands[terminal->commandCount++];
      int k;

      cmd->code = code;
      cmd->key = NOKEY;
      cmd->modifiers = 0;

      for (k=0; k<numkeys; k++) {
        int16_t key = keys[k];
        int found = 0;
        int m;
        for (m=0; m<terminal->modifierCount; m++) {
          if (key == terminal->modifiers[m]) {
            int bit = 1 << m;
            if (cmd->modifiers & bit) {
              yyerror("duplicate modifier");
              goto removeCommand;
            }
            cmd->modifiers |= bit;
            found = 1;
            break;
          }
        }

        if (!found) {
          if (cmd->key != NOKEY) {
            yyerror("more than one nonmodifier key");
            goto removeCommand;
          }
          cmd->key = key;
        }
      }
      return cmd;

    removeCommand:
      terminal->commandCount--;
    }
  }
  return NULL;
}
#line 380 "config.tab.c"
#define NUM 257
#define STRING 258
#define IS 259
#define AND 260
#define NAME 261
#define IDENT 262
#define HELPFILE 263
#define DISPLAYSIZE 264
#define STATCELLS 265
#define FRONTKEYS 266
#define HASBAR 267
#define MODIFIER 268
#define ROUTING1 269
#define ROUTING2 270
#define BAR 271
#define SWITCH 272
#define KEY 273
#define LEFT 274
#define RIGHT 275
#define REAR 276
#define FRONT 277
#define STAT 278
#define KEYCODE 279
#define STATCODE 280
#define BARCODE 281
#define NUMBER 282
#define FLAG 283
#define HORIZ 284
#define ON 285
#define OFF 286
#define CHECK 287
#define ROUTE 288
#define VPK 289
#define YYERRCODE 256
short yylhs[] = {                                        -1,
    0,    0,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    2,
    2,    2,    8,    8,    8,    3,    6,    6,    4,    4,
    4,    4,    5,    5,    5,    5,    5,    5,    5,    5,
    5,    5,    5,    5,    5,    7,    7,    9,
};
short yylen[] = {                                         2,
    0,    2,    1,    2,    4,    4,    4,    4,    4,    4,
    2,    7,    4,    4,    4,    5,    5,    5,    5,    1,
    1,    0,    1,    1,    0,    2,    1,    1,    1,    2,
    2,    2,    2,    2,    2,    3,    3,    3,    3,    3,
    3,    3,    3,    1,    1,    1,    3,    1,
};
short yydefred[] = {                                      1,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   27,   28,    3,    2,    0,    0,    4,   21,   20,
    0,    0,    0,    0,    0,    0,   11,    0,    0,   26,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   44,   45,    0,    0,    0,    0,    0,
    0,   29,    0,    0,    0,    0,    0,    0,    0,    0,
   48,    0,    0,    6,    5,    7,    8,    9,   10,    0,
   35,    0,    0,    0,    0,   34,   33,   14,   32,   31,
   30,   13,    0,    0,    0,    0,   15,   24,   23,    0,
    0,   36,   37,   38,   39,   40,   41,   42,   43,   16,
   17,   18,   19,   47,    0,   12,
};
short yydgoto[] = {                                       1,
   15,   21,   16,   56,   61,   17,   62,   90,   63,
};
short yysindex[] = {                                      0,
  -10,   16,  -56,  -56,  -56,  -56,  -56,  -56,   -9,  -56,
 -223,    0,    0,    0,    0,  -56,  -59,    0,    0,    0,
 -213, -195, -194, -192, -191, -190,    0, -189, -250,    0,
 -236,  -56,  -56,  -56,  -56, -250,   53,   59,   60,   61,
   62,   63, -183,    0,    0, -206, -264, -239, -181, -180,
   68,    0, -201, -200, -199,   72, -250, -250, -250, -250,
    0,   73,  -39,    0,    0,    0,    0,    0,    0, -173,
    0, -227, -220, -218, -216,    0,    0,    0,    0,    0,
    0,    0,   75,   76,   77,   78,    0,    0,    0, -250,
 -168,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   80,    0,
};
short yyrindex[] = {                                      0,
    0,    0, -167, -165, -167, -165, -165, -165, -165, -240,
    0,    0,    0,    0,    0, -229, -240,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, -240, -240, -240, -240,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   -7,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,
};
short yygindex[] = {                                      0,
    0,    8,    0,    0,   64,    0,  -51,    0,    0,
};
#define YYTABLESIZE 279
short yytable[] = {                                      14,
   27,   20,   46,   89,   20,   83,   84,   85,   86,   72,
   73,   22,   23,   24,   25,   26,   28,   29,   44,   45,
   46,   47,   48,   31,   36,   18,   49,   50,   22,   22,
   22,   22,   22,   30,   74,   75,   22,   22,  104,   57,
   58,   59,   60,   52,   37,   53,   54,   55,   92,   93,
   22,   20,   22,   22,   22,   94,   95,   96,   97,   98,
   99,   38,   64,   39,   40,   41,   42,   43,   65,   66,
   67,   68,   69,   70,   71,   76,   77,   78,   79,   80,
   81,   82,   87,   91,  100,  101,  102,  103,  105,  106,
   22,   22,   51,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   19,
    0,    0,   19,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   88,    0,    0,    0,    0,   32,   33,   34,   35,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    2,    0,    0,    0,   19,
    3,    4,    5,    6,    7,    8,    9,   10,    0,    0,
    0,   25,   25,   25,   25,   25,    0,   11,   12,   25,
   25,    0,    0,    0,    0,    0,    0,    0,   13,
};
short yycheck[] = {                                      10,
   10,   61,   10,   43,   61,   57,   58,   59,   60,  274,
  275,    4,    5,    6,    7,    8,    9,   10,  269,  270,
  271,  272,  273,   16,   17,   10,  277,  278,  269,  270,
  271,  272,  273,  257,  274,  275,  277,  278,   90,   32,
   33,   34,   35,  280,  258,  282,  283,  284,  276,  277,
  280,   61,  282,  283,  284,  276,  277,  276,  277,  276,
  277,  257,   10,  258,  257,  257,  257,  257,   10,   10,
   10,   10,   10,  257,  281,  257,  257,   10,  280,  280,
  280,   10,   10,  257,   10,   10,   10,   10,  257,   10,
  258,  257,   29,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  259,
   -1,   -1,  259,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  260,   -1,   -1,   -1,   -1,  285,  286,  287,  288,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  259,
  261,  262,  263,  264,  265,  266,  267,  268,   -1,   -1,
   -1,  269,  270,  271,  272,  273,   -1,  278,  279,  277,
  278,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  289,
};
#define YYFINAL 1
#ifndef YYDEBUG
#define YYDEBUG 0
#endif
#define YYMAXTOKEN 289
#if YYDEBUG
char *yyname[] = {
"end-of-file",0,0,0,0,0,0,0,0,0,"'\\n'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,"'+'",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"'='",0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"NUM","STRING","IS",
"AND","NAME","IDENT","HELPFILE","DISPLAYSIZE","STATCELLS","FRONTKEYS","HASBAR",
"MODIFIER","ROUTING1","ROUTING2","BAR","SWITCH","KEY","LEFT","RIGHT","REAR",
"FRONT","STAT","KEYCODE","STATCODE","BARCODE","NUMBER","FLAG","HORIZ","ON",
"OFF","CHECK","ROUTE","VPK",
};
char *yyrule[] = {
"$accept : input",
"input :",
"input : input inputline",
"inputline : '\\n'",
"inputline : error '\\n'",
"inputline : IDENT eq NUM '\\n'",
"inputline : NAME eq STRING '\\n'",
"inputline : HELPFILE eq STRING '\\n'",
"inputline : DISPLAYSIZE eq NUM '\\n'",
"inputline : STATCELLS eq NUM '\\n'",
"inputline : FRONTKEYS eq NUM '\\n'",
"inputline : HASBAR '\\n'",
"inputline : HASBAR eq NUM NUM NUM NUM '\\n'",
"inputline : statdef eq statdisp '\\n'",
"inputline : MODIFIER eq anykey '\\n'",
"inputline : keycode eq modifiers '\\n'",
"inputline : keycode ON eq modifiers '\\n'",
"inputline : keycode OFF eq modifiers '\\n'",
"inputline : keycode CHECK eq modifiers '\\n'",
"inputline : keycode ROUTE eq modifiers '\\n'",
"eq : '='",
"eq : IS",
"eq :",
"and : '+'",
"and : AND",
"and :",
"statdef : STAT NUM",
"keycode : KEYCODE",
"keycode : VPK",
"statdisp : STATCODE",
"statdisp : HORIZ STATCODE",
"statdisp : FLAG STATCODE",
"statdisp : NUMBER STATCODE",
"anykey : STAT NUM",
"anykey : FRONT NUM",
"anykey : BAR BARCODE",
"anykey : SWITCH LEFT REAR",
"anykey : SWITCH LEFT FRONT",
"anykey : SWITCH RIGHT REAR",
"anykey : SWITCH RIGHT FRONT",
"anykey : KEY LEFT REAR",
"anykey : KEY LEFT FRONT",
"anykey : KEY RIGHT REAR",
"anykey : KEY RIGHT FRONT",
"anykey : ROUTING1",
"anykey : ROUTING2",
"modifiers : modifier",
"modifiers : modifier and modifiers",
"modifier : anykey",
};
#endif
#ifndef YYSTYPE
typedef int YYSTYPE;
#endif
#if YYDEBUG
#include <stdio.h>
#endif

/* define the initial stack-sizes */
#ifdef YYSTACKSIZE
#undef YYMAXDEPTH
#define YYMAXDEPTH  YYSTACKSIZE
#else
#ifdef YYMAXDEPTH
#define YYSTACKSIZE YYMAXDEPTH
#else
#define YYSTACKSIZE 500
#define YYMAXDEPTH  500
#endif
#endif

#define YYINITSTACKSIZE 500

int      yydebug;
int      yynerrs;
int      yyerrflag;
int      yychar;
short   *yyssp;
YYSTYPE *yyvsp;
YYSTYPE  yyval;
YYSTYPE  yylval;

/* variables for the parser stack */
static short   *yyss;
static short   *yysslim;
static YYSTYPE *yyvs;
static int      yystacksize;
#line 444 "./config.y"

/* --------------------------------------------------------------- */
/* all the keywords */
/* the commands CMD_* CR_* in Programs/brl.h are autogenerated - see Makefile */

struct init_v
{
  char*  sname;			/* symbol (lowercase) */
  int token;			/* yacc token */
  int val;			/* key value */
};

static struct init_v symbols[]= {
  { "helpfile",    HELPFILE ,0 },
  { "is",          IS, 0 },
  { "and",         AND, 0 },

  { "identification", IDENT, 0 },
  { "identity",    IDENT, 0 },
  { "terminal",    NAME, 0 },
  { "type",        NAME, 0 },
  { "typ",         NAME, 0 },

  { "modifier",    MODIFIER, 0},

  { "displaysize", DISPLAYSIZE, 0 },
  { "statuscells", STATCELLS, 0 },
  { "frontkeys",   FRONTKEYS, 0 },
  { "hasbar",      HASBAR,    0 },

  { "status",      STAT,   0 },
  { "bar",         BAR,    0 },

  { "front",       FRONT,  0 },
  { "rear",        REAR,   0 },
  { "left",        LEFT,   0 },
  { "right",       RIGHT,  0 },

  { "switch",      SWITCH, 0 },
  { "key",         KEY,    0 },

  { "horiz",       HORIZ,  0 },
  { "flag",        FLAG,   0 },
  { "number",      NUMBER, 0 },

  { "left1",       BARCODE, BAR_L1},
  { "left2",       BARCODE, BAR_L2 },
  { "up1",         BARCODE, BAR_U1},
  { "up2",         BARCODE, BAR_U2 },
  { "right1",      BARCODE, BAR_R1},
  { "right2",      BARCODE, BAR_R2 },
  { "down1",       BARCODE, BAR_D1},
  { "down2",       BARCODE, BAR_D2 },

  { "routing1",    ROUTING1, 0 },
  { "routing2",    ROUTING2, 0 },

  { "on",          ON, 0 },
  { "off",         OFF, 0 },
  { "check",       CHECK, 0 },
  { "route",       ROUTE, 0 },

#include "cmd.auto.h"
  { "EOF", KEYCODE, -1 },
  { NULL, 0, 0 }
};

/* --------------------------------------------------------------- */
/* all the help */
/* the commands CMD_* in Programs/brl.h are autogenerated - see Makefile */

struct help_v
{
  int    cmd;			/* cmd */
  char*  txt;			/* help text */
};

static struct help_v hlptxt[]= {
#include "hlp.auto.h"
  { 0, NULL }
};


/* --------------------------------------------------------------- */

int yylex ()
{
  int c;
  static char symbuf[40] = { 0 };
  const int length = ARRAY_COUNT(symbuf);

  /* Ignore whitespace, get first nonwhite character.  */
  while ((c = getc(configurationFile)) == ' ' || c == '\t')
    ;

  if (c == EOF)
    return 0;

  if (c == '#') {		/* comment to end of line */
    do {
      c = getc(configurationFile);
    } while (c != '\n' && c != EOF);
    lineNumber++;
    return '\n';
  }

  /* Char starts a number => parse the number. */
  if (c == '.' || isdigit (c)) {
    ungetc (c, configurationFile);
    fscanf (configurationFile, "%d", &numval);
    yylval = numval;
    return NUM;
  }

  if (c == '"') {		/* string */
    int i=0;
    symbuf[0] = '\0';
    c = getc(configurationFile);
    while(c !='"' && c != EOF) {
      /* If buffer is full */
      if (i == length)
	break;
      /* Add this character to the buffer. */
      symbuf[i++] = c;
      c = getc(configurationFile);
    }
    symbuf[i] = 0;
    nameval = symbuf;
    return STRING;
  }

  /* Char starts an identifier => read the name. */
  if (isalpha (c) || c=='_') {
    int i = 0;
    do {
      /* If buffer is full */
      if (i == length)
	break;
      
      /* Add this character to the buffer. */
      symbuf[i++] = c;
      c = getc(configurationFile);
    } while (c != EOF && (isalnum (c) || c=='_'));
     
    ungetc (c, configurationFile);
    symbuf[i] = 0;
    for(i=0;symbols[i].sname; i++)
      if(strcasecmp(symbuf,symbols[i].sname) == 0) {
	numval = symbols[i].val;
	return symbols[i].token;
      }
    /* unused feature */
    nameval = symbuf;
    return NAME;
  }
     
  /* Any other character is a token by itself. */
  if (c == '\n')
    lineNumber++;

  return c;
}


/* --------------------------------------------------------------- */

void
deallocateTerminalTable (void) {
  while (pmTerminalCount) {
    TerminalDefinition *terminal = &pmTerminals[--pmTerminalCount];
    if (terminal->name) free(terminal->name);
    if (terminal->helpFile) free(terminal->helpFile);
    if (terminal->statusCells) free(terminal->statusCells);
    if (terminal->modifiers) free(terminal->modifiers);
    if (terminal->commands) free(terminal->commands);
  }

  if (pmTerminals) {
    free(pmTerminals);
    pmTerminals = NULL;
  }
}

int
parseConfigurationFile (void) {
  if (pmTerminalsAllocated) {
    deallocateTerminalTable();
  } else {
    pmTerminalCount = 0;
    pmTerminals = NULL;
    pmTerminalsAllocated = 1;
  }

  lineNumber = 1;
  terminalsSize = 0;

  nameval = NULL;
  numval = 0;
  keyindex = 0;
  cmdval = 0;
  numkeys = 0;

  {
    int result = yyparse();
    finishCurrentTerminal();
    return result;
  }
}
#line 851 "config.tab.c"
/* allocate initial stack or double stack size, up to YYMAXDEPTH */
static int yygrowstack(void)
{
    int newsize, i;
    short *newss;
    YYSTYPE *newvs;

    if ((newsize = yystacksize) == 0)
        newsize = YYINITSTACKSIZE;
    else if (newsize >= YYMAXDEPTH)
        return -1;
    else if ((newsize *= 2) > YYMAXDEPTH)
        newsize = YYMAXDEPTH;

    i = yyssp - yyss;
    newss = (yyss != 0)
          ? (short *)realloc(yyss, newsize * sizeof(*newss))
          : (short *)malloc(newsize * sizeof(*newss));
    if (newss == 0)
        return -1;

    yyss  = newss;
    yyssp = newss + i;
    newvs = (yyvs != 0)
          ? (YYSTYPE *)realloc(yyvs, newsize * sizeof(*newvs))
          : (YYSTYPE *)malloc(newsize * sizeof(*newvs));
    if (newvs == 0)
        return -1;

    yyvs = newvs;
    yyvsp = newvs + i;
    yystacksize = newsize;
    yysslim = yyss + newsize - 1;
    return 0;
}

#define YYABORT goto yyabort
#define YYREJECT goto yyabort
#define YYACCEPT goto yyaccept
#define YYERROR goto yyerrlab
int
yyparse(void)
{
    register int yym, yyn, yystate;
#if YYDEBUG
    register const char *yys;

    if ((yys = getenv("YYDEBUG")) != 0)
    {
        yyn = *yys;
        if (yyn >= '0' && yyn <= '9')
            yydebug = yyn - '0';
    }
#endif

    yynerrs = 0;
    yyerrflag = 0;
    yychar = YYEMPTY;

    if (yyss == NULL && yygrowstack()) goto yyoverflow;
    yyssp = yyss;
    yyvsp = yyvs;
    *yyssp = yystate = 0;

yyloop:
    if ((yyn = yydefred[yystate]) != 0) goto yyreduce;
    if (yychar < 0)
    {
        if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            printf("%sdebug: state %d, reading %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
    }
    if ((yyn = yysindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
#if YYDEBUG
        if (yydebug)
            printf("%sdebug: state %d, shifting to state %d\n",
                    YYPREFIX, yystate, yytable[yyn]);
#endif
        if (yyssp >= yysslim && yygrowstack())
        {
            goto yyoverflow;
        }
        *++yyssp = yystate = yytable[yyn];
        *++yyvsp = yylval;
        yychar = YYEMPTY;
        if (yyerrflag > 0)  --yyerrflag;
        goto yyloop;
    }
    if ((yyn = yyrindex[yystate]) && (yyn += yychar) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yychar)
    {
        yyn = yytable[yyn];
        goto yyreduce;
    }
    if (yyerrflag) goto yyinrecovery;

    yyerror("syntax error");

#ifdef lint
    goto yyerrlab;
#endif

yyerrlab:
    ++yynerrs;

yyinrecovery:
    if (yyerrflag < 3)
    {
        yyerrflag = 3;
        for (;;)
        {
            if ((yyn = yysindex[*yyssp]) && (yyn += YYERRCODE) >= 0 &&
                    yyn <= YYTABLESIZE && yycheck[yyn] == YYERRCODE)
            {
#if YYDEBUG
                if (yydebug)
                    printf("%sdebug: state %d, error recovery shifting\
 to state %d\n", YYPREFIX, *yyssp, yytable[yyn]);
#endif
                if (yyssp >= yysslim && yygrowstack())
                {
                    goto yyoverflow;
                }
                *++yyssp = yystate = yytable[yyn];
                *++yyvsp = yylval;
                goto yyloop;
            }
            else
            {
#if YYDEBUG
                if (yydebug)
                    printf("%sdebug: error recovery discarding state %d\n",
                            YYPREFIX, *yyssp);
#endif
                if (yyssp <= yyss) goto yyabort;
                --yyssp;
                --yyvsp;
            }
        }
    }
    else
    {
        if (yychar == 0) goto yyabort;
#if YYDEBUG
        if (yydebug)
        {
            yys = 0;
            if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
            if (!yys) yys = "illegal-symbol";
            printf("%sdebug: state %d, error recovery discards token %d (%s)\n",
                    YYPREFIX, yystate, yychar, yys);
        }
#endif
        yychar = YYEMPTY;
        goto yyloop;
    }

yyreduce:
#if YYDEBUG
    if (yydebug)
        printf("%sdebug: state %d, reducing by rule %d (%s)\n",
                YYPREFIX, yystate, yyn, yyrule[yyn]);
#endif
    yym = yylen[yyn];
    yyval = yyvsp[1-yym];
    switch (yyn)
    {
case 4:
#line 379 "./config.y"
{ yyerrok;  }
break;
case 5:
#line 380 "./config.y"
{ addTerminal(numval); }
break;
case 6:
#line 381 "./config.y"
{ setName(nameval); }
break;
case 7:
#line 382 "./config.y"
{ setHelp(nameval); }
break;
case 8:
#line 383 "./config.y"
{ setColumns(numval); }
break;
case 9:
#line 384 "./config.y"
{ setStatusCells(numval); }
break;
case 10:
#line 385 "./config.y"
{ setFrontKeys(numval); }
break;
case 11:
#line 386 "./config.y"
{ setHasBar(1, 1, 1, 1); }
break;
case 12:
#line 387 "./config.y"
{ setHasBar(yyvsp[-4], yyvsp[-3], yyvsp[-2], yyvsp[-1]); }
break;
case 13:
#line 389 "./config.y"
{ setStatusCell(keyindex, numval);  }
break;
case 14:
#line 390 "./config.y"
{ addModifier(keyindex); }
break;
case 15:
#line 391 "./config.y"
{ addCommand(cmdval); }
break;
case 16:
#line 392 "./config.y"
{ addCommand(cmdval | BRL_FLG_TOGGLE_ON); }
break;
case 17:
#line 393 "./config.y"
{ addCommand(cmdval | BRL_FLG_TOGGLE_OFF); }
break;
case 18:
#line 394 "./config.y"
{ addCommand(cmdval | BRL_FLG_TOGGLE_ON | BRL_FLG_TOGGLE_OFF); }
break;
case 19:
#line 395 "./config.y"
{ addCommand(cmdval | BRL_FLG_ROUTE); }
break;
case 26:
#line 408 "./config.y"
{ keyindex=numval;  }
break;
case 27:
#line 411 "./config.y"
{ cmdval=numval; numkeys = 0; }
break;
case 28:
#line 412 "./config.y"
{ cmdval = numval; numkeys = 0; }
break;
case 29:
#line 415 "./config.y"
{  }
break;
case 30:
#line 416 "./config.y"
{ numval += OFFS_HORIZ; }
break;
case 31:
#line 417 "./config.y"
{ numval += OFFS_FLAG; }
break;
case 32:
#line 418 "./config.y"
{ numval += OFFS_NUMBER; }
break;
case 33:
#line 421 "./config.y"
{ keyindex= OFFS_STAT + numval; }
break;
case 34:
#line 422 "./config.y"
{ keyindex= OFFS_FRONT + numval; }
break;
case 35:
#line 423 "./config.y"
{ keyindex= OFFS_BAR + numval; }
break;
case 36:
#line 424 "./config.y"
{ keyindex= OFFS_SWITCH + SWITCH_LEFT_REAR; }
break;
case 37:
#line 425 "./config.y"
{ keyindex= OFFS_SWITCH + SWITCH_LEFT_FRONT; }
break;
case 38:
#line 426 "./config.y"
{ keyindex= OFFS_SWITCH + SWITCH_RIGHT_REAR; }
break;
case 39:
#line 427 "./config.y"
{ keyindex= OFFS_SWITCH + SWITCH_RIGHT_FRONT; }
break;
case 40:
#line 428 "./config.y"
{ keyindex= OFFS_SWITCH + KEY_LEFT_REAR; }
break;
case 41:
#line 429 "./config.y"
{ keyindex= OFFS_SWITCH + KEY_LEFT_FRONT; }
break;
case 42:
#line 430 "./config.y"
{ keyindex= OFFS_SWITCH + KEY_RIGHT_REAR; }
break;
case 43:
#line 431 "./config.y"
{ keyindex= OFFS_SWITCH + KEY_RIGHT_FRONT; }
break;
case 44:
#line 432 "./config.y"
{ keyindex= ROUTINGKEY1; }
break;
case 45:
#line 433 "./config.y"
{ keyindex= ROUTINGKEY2; }
break;
case 48:
#line 440 "./config.y"
{ keys[numkeys++] = keyindex; }
break;
#line 1177 "config.tab.c"
    }
    yyssp -= yym;
    yystate = *yyssp;
    yyvsp -= yym;
    yym = yylhs[yyn];
    if (yystate == 0 && yym == 0)
    {
#if YYDEBUG
        if (yydebug)
            printf("%sdebug: after reduction, shifting from state 0 to\
 state %d\n", YYPREFIX, YYFINAL);
#endif
        yystate = YYFINAL;
        *++yyssp = YYFINAL;
        *++yyvsp = yyval;
        if (yychar < 0)
        {
            if ((yychar = yylex()) < 0) yychar = 0;
#if YYDEBUG
            if (yydebug)
            {
                yys = 0;
                if (yychar <= YYMAXTOKEN) yys = yyname[yychar];
                if (!yys) yys = "illegal-symbol";
                printf("%sdebug: state %d, reading %d (%s)\n",
                        YYPREFIX, YYFINAL, yychar, yys);
            }
#endif
        }
        if (yychar == 0) goto yyaccept;
        goto yyloop;
    }
    if ((yyn = yygindex[yym]) && (yyn += yystate) >= 0 &&
            yyn <= YYTABLESIZE && yycheck[yyn] == yystate)
        yystate = yytable[yyn];
    else
        yystate = yydgoto[yym];
#if YYDEBUG
    if (yydebug)
        printf("%sdebug: after reduction, shifting from state %d \
to state %d\n", YYPREFIX, *yyssp, yystate);
#endif
    if (yyssp >= yysslim && yygrowstack())
    {
        goto yyoverflow;
    }
    *++yyssp = yystate;
    *++yyvsp = yyval;
    goto yyloop;

yyoverflow:
    yyerror("yacc stack overflow");

yyabort:
    return (1);

yyaccept:
    return (0);
}
